/*
 * Copyright 2002-2006 the original author or authors.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package net.sf.json;

/*
Copyright (c) 2002 JSON.org

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

The Software shall be used for Good, not Evil.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
*/


import java.beans.PropertyDescriptor;
import java.io.IOException;
import java.io.Writer;
import java.lang.reflect.Array;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import net.sf.json.regexp.RegexpUtils;
import net.sf.json.util.JSONTokener;
import net.sf.json.util.JSONUtils;

import org.apache.commons.beanutils.DynaBean;
import org.apache.commons.beanutils.DynaProperty;
import org.apache.commons.beanutils.PropertyUtils;
import org.apache.commons.lang.StringUtils;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

/**
 * A JSONObject is an unordered collection of name/value pairs. Its external
 * form is a string wrapped in curly braces with colons between the names and
 * values, and commas between the values and names. The internal form is an
 * object having <code>get</code> and <code>opt</code> methods for accessing
 * the values by name, and <code>put</code> methods for adding or replacing
 * values by name. The values can be any of these types: <code>Boolean</code>,
 * <code>JSONArray</code>, <code>JSONObject</code>, <code>Number</code>,
 * <code>String</code>, or the <code>JSONNull</code> object. A JSONObject
 * constructor can be used to convert an external form JSON text into an
 * internal form whose values can be retrieved with the <code>get</code> and
 * <code>opt</code> methods, or to convert values into a JSON text using the
 * <code>put</code> and <code>toString</code> methods. A <code>get</code>
 * method returns a value if one can be found, and throws an exception if one
 * cannot be found. An <code>opt</code> method returns a default value instead
 * of throwing an exception, and so is useful for obtaining optional values.
 * <p>
 * The generic <code>get()</code> and <code>opt()</code> methods return an
 * object, which you can cast or query for type. There are also typed
 * <code>get</code> and <code>opt</code> methods that do type checking and
 * type coersion for you.
 * <p>
 * The <code>put</code> methods adds values to an object. For example,
 *
 * <pre>
 *     myString = new JSONObject().put("JSON", "Hello, World!").toString();</pre>
 *
 * produces the string <code>{"JSON": "Hello, World"}</code>.
 * <p>
 * The texts produced by the <code>toString</code> methods strictly conform to
 * the JSON sysntax rules. The constructors are more forgiving in the texts they
 * will accept:
 * <ul>
 * <li>An extra <code>,</code>&nbsp;<small>(comma)</small> may appear just
 * before the closing brace.</li>
 * <li>Strings may be quoted with <code>'</code>&nbsp;<small>(single quote)</small>.</li>
 * <li>Strings do not need to be quoted at all if they do not begin with a
 * quote or single quote, and if they do not contain leading or trailing spaces,
 * and if they do not contain any of these characters:
 * <code>{ } [ ] / \ : , = ; #</code> and if they do not look like numbers and
 * if they are not the reserved words <code>true</code>, <code>false</code>,
 * or <code>null</code>.</li>
 * <li>Keys can be followed by <code>=</code> or <code>=></code> as well as
 * by <code>:</code>.</li>
 * <li>Values can be followed by <code>;</code> <small>(semicolon)</small>
 * as well as by <code>,</code> <small>(comma)</small>.</li>
 * <li>Numbers may have the <code>0-</code> <small>(octal)</small> or
 * <code>0x-</code> <small>(hex)</small> prefix.</li>
 * <li>Comments written in the slashshlash, slashstar, and hash conventions
 * will be ignored.</li>
 * </ul>
 *
 * @author JSON.org
 * @version 5
 */
public final class JSONObject implements JSON
{
   private static final String[] defaultExcludes = new String[] { "class" };

   private static final Log log = LogFactory.getLog( JSONObject.class );

   /**
    * Creates a JSONObject from a POJO.<br>
    * Supports nested maps, POJOs, and arrays/collections.
    *
    * @param bean An object with POJO conventions
    * @throws JSONException if the bean can not be converted to a proper
    *         JSONObject
    */
   public static JSONObject fromBean( Object bean )
   {
      return fromBean( bean, null, false );
   }

   /**
    * Creates a JSONObject from a POJO.<br>
    * Supports nested maps, POJOs, and arrays/collections.
    *
    * @param bean An object with POJO conventions
    * @param excludes A group of property names to be excluded
    * @throws JSONException if the bean can not be converted to a proper
    *         JSONObject
    */
   public static JSONObject fromBean( Object bean, String[] excludes )
   {
      return fromBean( bean, excludes, false );
   }

   /**
    * Creates a JSONObject from a POJO.<br>
    * Supports nested maps, POJOs, and arrays/collections.
    *
    * @param bean An object with POJO conventions
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @throws JSONException if the bean can not be converted to a proper
    *         JSONObject.
    */
   public static JSONObject fromBean( Object bean, String[] excludes, boolean ignoreDefaultExcludes )
   {
      if( bean == null || JSONUtils.isNull( bean ) ){
         return new JSONObject( true );
      }else if( bean instanceof JSONObject ){
         return fromJSONObject( (JSONObject) bean, excludes, ignoreDefaultExcludes );
      }else if( bean instanceof DynaBean ){
         return fromDynaBean( (DynaBean) bean, excludes, ignoreDefaultExcludes );
      }else if( bean instanceof JSONTokener ){
         return fromJSONTokener( (JSONTokener) bean, excludes, ignoreDefaultExcludes );
      }else if( bean instanceof JSONString ){
         return fromJSONString( (JSONString) bean, excludes, ignoreDefaultExcludes );
      }else if( bean instanceof Map ){
         return fromMap( (Map) bean, excludes, ignoreDefaultExcludes );
      }else if( bean instanceof String ){
         return fromString( (String) bean, excludes, ignoreDefaultExcludes );
      }else if( JSONUtils.isNumber( bean ) || JSONUtils.isBoolean( bean )
            || JSONUtils.isString( bean ) ){
         return new JSONObject();
      }else if( JSONUtils.isArray( bean ) ){
         throw new JSONException( "'bean' is an array. Use JSONArray instead" );
      }else{
         JSONObject jsonObject = new JSONObject();
         Collection exclusions = mergeExclusions( excludes, ignoreDefaultExcludes );
         try{
            PropertyDescriptor[] pds = PropertyUtils.getPropertyDescriptors( bean );
            for( int i = 0; i < pds.length; i++ ){
               String key = pds[i].getName();
               if( exclusions.contains( key ) ){
                  continue;
               }

               Class type = pds[i].getPropertyType();
               if( pds[i].getReadMethod() != null ){
                  Object value = PropertyUtils.getProperty( bean, key );
                  setValue( jsonObject, key, value, type, excludes, ignoreDefaultExcludes );
               }else{
                  log.warn( "Property '" + key + "' has no read method. SKIPPED" );
               }
            }
         }
         catch( JSONException jsone ){
            throw jsone;
         }
         catch( Exception e ){
            throw new JSONException( e );
         }
         return jsonObject;
      }
   }

   /**
    * Creates a JSONObject from a DynaBean.<br>
    * Supports nested maps, POJOs, and arrays/collections.
    *
    * @param bean A DynaBean implementation
    * @throws JSONException if the bean can not be converted to a proper
    *         JSONObject
    */
   public static JSONObject fromDynaBean( DynaBean bean )
   {
      return fromDynaBean( bean, null, false );
   }

   /**
    * Creates a JSONObject from a DynaBean.<br>
    * Supports nested maps, POJOs, and arrays/collections.
    *
    * @param bean A DynaBean implementation
    * @param excludes A group of property names to be excluded
    * @throws JSONException if the bean can not be converted to a proper
    *         JSONObject
    */
   public static JSONObject fromDynaBean( DynaBean bean, String[] excludes )
   {
      return fromDynaBean( bean, excludes, false );
   }

   /**
    * Creates a JSONObject from a DynaBean.<br>
    * Supports nested maps, POJOs, and arrays/collections.
    *
    * @param bean A DynaBean implementation
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @throws JSONException if the bean can not be converted to a proper
    *         JSONObject
    */
   public static JSONObject fromDynaBean( DynaBean bean, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      return new JSONObject( bean, excludes, ignoreDefaultExcludes );
   }

   /**
    * Constructs a JSONObject from another JSONObject.
    */
   public static JSONObject fromJSONObject( JSONObject object )
   {
      return fromJSONObject( object, null, false );
   }

   /**
    * Constructs a JSONObject from another JSONObject. *
    *
    * @param excludes A group of property names to be excluded
    */
   public static JSONObject fromJSONObject( JSONObject object, String[] excludes )
   {
      return fromJSONObject( object, excludes, false );
   }

   /**
    * Constructs a JSONObject from another JSONObject. *
    *
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    */
   public static JSONObject fromJSONObject( JSONObject object, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      return new JSONObject( object, excludes, ignoreDefaultExcludes );
   }

   /**
    * Creates a JSONObject from a JSONString.<br>
    *
    * @param string A string with JSON values
    * @throws JSONException if the source string is not a valid JSON string for
    *         a JSONObject
    */
   public static JSONObject fromJSONString( JSONString string )
   {
      return fromJSONTokener( new JSONTokener( string.toJSONString() ), null, false );
   }

   /**
    * Creates a JSONObject from a JSONString.<br>
    *
    * @param string A string with JSON values
    * @param excludes A group of property names to be excluded
    * @throws JSONException if the source string is not a valid JSON string for
    *         a JSONObject
    */
   public static JSONObject fromJSONString( JSONString string, String[] excludes )
   {
      return fromJSONTokener( new JSONTokener( string.toJSONString() ), excludes, false );
   }

   /**
    * Creates a JSONObject from a JSONString.<br>
    *
    * @param string A string with JSON values
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @throws JSONException if the source string is not a valid JSON string for
    *         a JSONObject
    */
   public static JSONObject fromJSONString( JSONString string, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      return fromJSONTokener( new JSONTokener( string.toJSONString() ), excludes,
            ignoreDefaultExcludes );
   }

   /**
    * Creates a JSONObject from a map.<br>
    * The key names will become the object's attributes. Supports nested maps,
    * POJOs, and arrays/collections.
    *
    * @param map
    * @throws JSONException if the map contains invalid JSON values
    */
   public static JSONObject fromMap( Map map )
   {
      return fromMap( map, null, false );
   }

   /**
    * Creates a JSONObject from a map.<br>
    * The key names will become the object's attributes. Supports nested maps,
    * POJOs, and arrays/collections.
    *
    * @param map
    * @param excludes A group of property names to be excluded
    * @throws JSONException if the map contains invalid JSON values
    */
   public static JSONObject fromMap( Map map, String[] excludes )
   {
      return fromMap( map, excludes, false );
   }

   /**
    * Creates a JSONObject from a map.<br>
    * The key names will become the object's attributes. Supports nested maps,
    * POJOs, and arrays/collections.
    *
    * @param map
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @throws JSONException if the map contains invalid JSON values
    */
   public static JSONObject fromMap( Map map, String[] excludes, boolean ignoreDefaultExcludes )
   {
      return new JSONObject( map, excludes, ignoreDefaultExcludes );
   }

   /**
    * Creates a JSONObject.<br>
    * Inspects the object type to call the correct JSONObject factory method.
    *
    * @param object
    * @throws JSONException if the object can not be converted to a proper
    *         JSONObject.
    */
   public static JSONObject fromObject( Object object )
   {
      return fromObject( object, null, false );
   }

   /**
    * Creates a JSONObject.<br>
    * Inspects the object type to call the correct JSONObject factory method.
    *
    * @param object
    * @param excludes A group of property names to be excluded
    * @throws JSONException if the object can not be converted to a proper
    *         JSONObject.
    */
   public static JSONObject fromObject( Object object, String[] excludes )
   {
      return fromObject( object, excludes, false );
   }

   /**
    * Creates a JSONObject.<br>
    * Inspects the object type to call the correct JSONObject factory method.
    *
    * @param object
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @throws JSONException if the object can not be converted to a proper
    *         JSONObject.
    */
   public static JSONObject fromObject( Object object, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      if( object == null || JSONUtils.isNull( object ) ){
         return new JSONObject( true );
      }else if( object instanceof JSONObject ){
         return fromJSONObject( (JSONObject) object, excludes, ignoreDefaultExcludes );
      }else if( object instanceof DynaBean ){
         return fromDynaBean( (DynaBean) object, excludes, ignoreDefaultExcludes );
      }else if( object instanceof JSONTokener ){
         return fromJSONTokener( (JSONTokener) object, excludes, ignoreDefaultExcludes );
      }else if( object instanceof JSONString ){
         return fromJSONString( (JSONString) object, excludes, ignoreDefaultExcludes );
      }else if( object instanceof Map ){
         return fromMap( (Map) object, excludes, ignoreDefaultExcludes );
      }else if( object instanceof String ){
         return fromString( (String) object, excludes, ignoreDefaultExcludes );
      }else if( JSONUtils.isNumber( object ) || JSONUtils.isBoolean( object )
            || JSONUtils.isString( object ) ){
         return new JSONObject();
      }else if( JSONUtils.isArray( object ) ){
         throw new JSONException( "'object' is an array. Use JSONArray instead" );
      }else{
         return fromBean( object, excludes, ignoreDefaultExcludes );
      }
   }

   /**
    * Constructs a JSONObject from a string in JSON format.
    *
    * @param str A string in JSON format
    * @throws JSONException if the source string is not a valid JSON string for
    *         a JSONObject
    */
   public static JSONObject fromString( String str )
   {
      return fromString( str, null, false );
   }

   /**
    * Constructs a JSONObject from a string in JSON format.
    *
    * @param str A string in JSON format
    * @param excludes A group of property names to be excluded
    * @throws JSONException if the source string is not a valid JSON string for
    *         a JSONObject
    */
   public static JSONObject fromString( String str, String[] excludes )
   {
      return fromString( str, excludes, false );
   }

   /**
    * Constructs a JSONObject from a string in JSON format.
    *
    * @param str A string in JSON format
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @throws JSONException if the source string is not a valid JSON string for
    *         a JSONObject
    */
   public static JSONObject fromString( String str, String[] excludes, boolean ignoreDefaultExcludes )
   {
      if( str == null || "null".compareToIgnoreCase( str ) == 0 ){
         return new JSONObject( true );
      }
      return fromJSONTokener( new JSONTokener( str ), excludes, ignoreDefaultExcludes );
   }

   /**
    * Creates a JSONDynaBean from a JSONObject.
    */
   public static Object toBean( JSONObject jsonObject )
   {
      if( jsonObject == null || jsonObject.isNullObject() ){
         return null;
      }

      DynaBean dynaBean = null;
      try{
         Map props = JSONUtils.getProperties( jsonObject );
         dynaBean = JSONUtils.newDynaBean( jsonObject );
         for( Iterator entries = jsonObject.names()
               .iterator(); entries.hasNext(); ){
            String name = (String) entries.next();
            String key = JSONUtils.convertToJavaIdentifier( name );
            Class type = (Class) props.get( key );
            Object value = jsonObject.get( name );

            if( !JSONUtils.isNull( value ) ){
               if( value instanceof JSONArray ){
                  setProperty( dynaBean, key, JSONArray.toList( (JSONArray) value ) );
               }else if( String.class.isAssignableFrom( type )
                     || Boolean.class.isAssignableFrom( type ) || JSONUtils.isNumber( type )
                     || Character.class.isAssignableFrom( type )
                     || JSONFunction.class.isAssignableFrom( type ) ){
                  setProperty( dynaBean, key, value );
               }else{
                  setProperty( dynaBean, key, toBean( (JSONObject) value ) );
               }
            }else{
               if( type.isPrimitive() ){
                  // assume assigned default value
                  log.warn( "Tried to assign null value to " + key + ":" + type.getName() );
                  setProperty( dynaBean, key, JSONUtils.getMorpherRegistry()
                        .morph( type, null ) );
               }else{
                  setProperty( dynaBean, key, null );
               }
            }
         }
      }
      catch( JSONException jsone ){
         throw jsone;
      }
      catch( Exception e ){
         throw new JSONException( e );
      }

      return dynaBean;
   }

   /**
    * Creates a bean from a JSONObject, with a specific target class.<br>
    */
   public static Object toBean( JSONObject jsonObject, Class beanClass )
   {
      return toBean( jsonObject, beanClass, null );
   }

   /**
    * Creates a bean from a JSONObject, with a specific target class.<br>
    * If beanClass is null, this method will return a graph of DynaBeans. Any
    * attribute that is a JSONObject and matches a key in the classMap will be
    * converted to that target class.<br>
    * The classMap has the following conventions:
    * <ul>
    * <li>Every key must be an String.</li>
    * <li>Every value must be a Class.</li>
    * <li>A key may be a regular expression.</li>
    * </ul>
    */
   public static Object toBean( JSONObject jsonObject, Class beanClass, Map classMap )
   {
      if( jsonObject == null || jsonObject.isNullObject() ){
         return null;
      }

      if( beanClass == null ){
         return toBean( jsonObject );
      }
      if( classMap == null ){
         classMap = Collections.EMPTY_MAP;
      }

      Object bean = null;
      try{
         if( beanClass.isInterface() ){
            if( !Map.class.isAssignableFrom( beanClass ) ){
               throw new JSONException( "beanClass is an interface. " + beanClass );
            }else{
               bean = new HashMap();
            }
         }else{
            bean = beanClass.newInstance();
         }
         Map props = JSONUtils.getProperties( jsonObject );
         for( Iterator entries = jsonObject.names()
               .iterator(); entries.hasNext(); ){
            String name = (String) entries.next();
            String key = JSONUtils.convertToJavaIdentifier( name );
            Class type = (Class) props.get( key );
            Object value = jsonObject.get( name );

            PropertyDescriptor pd = PropertyUtils.getPropertyDescriptor( bean, key );
            if( pd != null && pd.getWriteMethod() == null ){
               log.warn( "Property '" + key + "' has no write method. SKIPPED." );
               continue;
            }

            if( !JSONUtils.isNull( value ) ){
               if( value instanceof JSONArray ){
                  if( List.class.isAssignableFrom( pd.getPropertyType() ) ){
                     Class targetClass = findTargetClass( key, classMap );
                     targetClass = targetClass == null ? findTargetClass( name, classMap )
                           : targetClass;
                     // if targetClass is null the outcome will be a List of
                     // DynaBeans
                     // targetClass = (targetClass != null) ? targetClass :
                     // beanClass;
                     List list = JSONArray.toList( (JSONArray) value, targetClass, classMap );
                     setProperty( bean, key, list );
                  }else{
                     Object array = JSONArray.toArray( (JSONArray) value, beanClass, classMap );
                     Class innerType = JSONUtils.getInnerComponentType( pd.getPropertyType() );
                     if( innerType.isPrimitive() || JSONUtils.isNumber( innerType )
                           || Boolean.class.isAssignableFrom( innerType )
                           || JSONUtils.isString( innerType ) || !array.getClass()
                                 .equals( pd.getPropertyType() ) ){
                        array = JSONUtils.getMorpherRegistry()
                              .morph( Array.newInstance( innerType, 0 )
                                    .getClass(), array );
                     }
                     setProperty( bean, key, array );
                  }
               }else if( String.class.isAssignableFrom( type ) || JSONUtils.isBoolean( type )
                     || JSONUtils.isNumber( type ) || JSONUtils.isString( type )
                     || JSONFunction.class.isAssignableFrom( type ) ){
                  if( pd != null ){
                     if( !pd.getPropertyType()
                           .isInstance( value ) ){
                        setProperty( bean, key, JSONUtils.getMorpherRegistry()
                              .morph( pd.getPropertyType(), value ) );
                     }else{
                        setProperty( bean, key, value );
                     }
                  }else{
                     setProperty( bean, key, value );
                  }
               }else{
                  if( pd != null ){
                     Class targetClass = pd.getPropertyType();
                     if( targetClass == Object.class ){
                        targetClass = findTargetClass( key, classMap );
                        targetClass = targetClass == null ? findTargetClass( name, classMap )
                              : targetClass;
                     }
                     setProperty( bean, key, toBean( (JSONObject) value, targetClass, classMap ) );
                  }else{
                     Class targetClass = findTargetClass( key, classMap );
                     targetClass = targetClass == null ? findTargetClass( name, classMap )
                           : targetClass;
                     if( targetClass != null ){
                        setProperty( bean, key, toBean( (JSONObject) value, targetClass, classMap ) );
                     }else{
                        setProperty( bean, key, toBean( (JSONObject) value ) );
                     }
                  }
               }
            }else{
               if( type.isPrimitive() ){
                  // assume assigned default value
                  log.warn( "Tried to assign null value to " + key + ":" + type.getName() );
                  setProperty( bean, key, JSONUtils.getMorpherRegistry()
                        .morph( type, null ) );
               }else{
                  setProperty( bean, key, null );
               }
            }
         }
      }
      catch( JSONException jsone ){
         throw jsone;
      }
      catch( Exception e ){
         throw new JSONException( e );
      }

      return bean;
   }

   /**
    * Locates a Class associated to a specifi key.<br>
    * The key may be a regexp.
    */
   private static Class findTargetClass( String key, Map classMap )
   {
      // try get first
      Class targetClass = (Class) classMap.get( key );
      if( targetClass == null ){
         // try with regexp
         // this will hit performance as it must iterate over all the keys
         // and create a RegexpMatcher for each key
         for( Iterator i = classMap.entrySet()
               .iterator(); i.hasNext(); ){
            Map.Entry entry = (Map.Entry) i.next();
            if( RegexpUtils.getMatcher( (String) entry.getKey() )
                  .matches( key ) ){
               targetClass = (Class) entry.getValue();
               break;
            }
         }
      }

      return targetClass;
   }

   /**
    * Constructs a JSONObject from a JSONTokener.
    *
    * @param tokener
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @throws JSONException if the tokener can not correctly process its input
    */
   private static JSONObject fromJSONTokener( JSONTokener tokener, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      return new JSONObject( tokener, excludes, ignoreDefaultExcludes );
   }

   /**
    * Merge excludes and defualtExcludes if needed.
    */
   private static Collection mergeExclusions( String[] excludes, boolean ignoreDefaultExcludes )
   {
      Collection exclusions = new HashSet();
      if( excludes != null ){
         for( int i = 0; i < excludes.length; i++ ){
            String exclusion = excludes[i];
            if( excludes[i] != null && exclusion.trim()
                  .length() > 0 ){
               exclusions.add( exclusion.trim() );
            }
         }
      }
      if( !ignoreDefaultExcludes ){
         for( int i = 0; i < defaultExcludes.length; i++ ){
            if( !exclusions.contains( defaultExcludes[i] ) ){
               exclusions.add( defaultExcludes[i] );
            }
         }
      }

      return exclusions;
   }

   /**
    * Sets a property on the target bean.<br>
    * Bean may be a Map, a JSONObject, a DynaBean or a POJO.
    */
   private static void setProperty( Object bean, String key, Object value ) throws Exception
   {
      setProperty( bean, key, value, null, false );
   }

   /**
    * Sets a property on the target bean.<br>
    * Bean may be a Map, a JSONObject, a DynaBean or a POJO.
    */
   private static void setProperty( Object bean, String key, Object value, String[] excludes,
         boolean ignoreDefaultExcludes ) throws Exception
   {
      if( bean instanceof Map ){
         ((Map) bean).put( key, value );
      }else if( bean instanceof JSONObject ){
         if( !((JSONObject) bean).isNullObject() ){
            ((JSONObject) bean).put( key, value, excludes, ignoreDefaultExcludes );
         }
      }else{
         if( !JSONUtils.isJavaIdentifier( key ) ){
            key = JSONUtils.convertToJavaIdentifier( key );
         }
         PropertyUtils.setProperty( bean, key, value );
      }
   }

   /**
    * Sets a property on the target bean, converting it to a valid JSON value.
    */
   private static void setValue( Object object, String key, Object value, Class type,
         String[] excludes, boolean ignoreDefaultExcludes )
   {
      if( key == null ){
         throw new JSONException( "Null key" );
      }
      try{
         if( (Class.class.isAssignableFrom( type ) && value != null) || value instanceof Class ){
            setProperty( object, key, ((Class) value).getName(), excludes, ignoreDefaultExcludes );
         }else if( JSONUtils.isFunction( value ) ){
            setProperty( object, key, value, excludes, ignoreDefaultExcludes );
         }else if( value instanceof JSONString ){
            setProperty( object, key, JSONSerializer.toJSON( (JSONString) value, excludes,
                  ignoreDefaultExcludes ), excludes, ignoreDefaultExcludes );
         }else if( JSONUtils.isArray( value ) ){
            setProperty( object, key,
                  JSONArray.fromObject( value, excludes, ignoreDefaultExcludes ), excludes,
                  ignoreDefaultExcludes );
         }else if( value instanceof JSON ){
            setProperty( object, key, value, excludes, ignoreDefaultExcludes );
         }else if( String.class.isAssignableFrom( type ) || JSONUtils.isString( value ) ){
            String str = String.valueOf( value );
            if( value == null ){
               setProperty( object, key, "", excludes, ignoreDefaultExcludes );
            }else if( JSONUtils.mayBeJSON( str ) ){
               try{
                  setProperty( object, key, JSONSerializer.toJSON( str, excludes,
                        ignoreDefaultExcludes ), excludes, ignoreDefaultExcludes );
               }
               catch( JSONException jsone ){
                  setProperty( object, key, JSONUtils.stripQuotes( str ), excludes,
                        ignoreDefaultExcludes );
               }
            }else{
               setProperty( object, key, JSONUtils.stripQuotes( str ), excludes,
                     ignoreDefaultExcludes );
            }
         }else if( JSONUtils.isNumber( value ) ){
            JSONUtils.testValidity( value );
            setProperty( object, key, value, excludes, ignoreDefaultExcludes );
         }else if( JSONUtils.isBoolean( value ) ){
            setProperty( object, key, value, excludes, ignoreDefaultExcludes );
         }else if( value == null ){
            if( JSONUtils.isArray( type ) ){
               setProperty( object, key, JSONSerializer.toJSON( "[]" ), excludes,
                     ignoreDefaultExcludes );
            }else if( JSONUtils.isNumber( type ) ){
               if( JSONUtils.isDouble( type ) ){
                  setProperty( object, key, new Double( 0 ), excludes, ignoreDefaultExcludes );
               }else{
                  setProperty( object, key, new Integer( 0 ), excludes, ignoreDefaultExcludes );
               }
            }else if( JSONUtils.isBoolean( type ) ){
               setProperty( object, key, "false", excludes, ignoreDefaultExcludes );
            }else if( JSONUtils.isString( type ) ){
               setProperty( object, key, "", excludes, ignoreDefaultExcludes );
            }else{
               setProperty( object, key, JSONNull.getInstance(), excludes, ignoreDefaultExcludes );
            }
         }else{
            setProperty( object, key, fromObject( value, excludes, ignoreDefaultExcludes ),
                  excludes, ignoreDefaultExcludes );
         }
      }
      catch( JSONException jsone ){
         throw jsone;
      }
      catch( Exception e ){
         throw new JSONException( e );
      }
   }
   // ------------------------------------------------------

   /** identifies this object as null */
   private boolean nullObject;

   /**
    * The Map where the JSONObject's properties are kept.
    */
   private Map properties;

   /**
    * Construct an empty JSONObject.
    */
   public JSONObject()
   {
      this.properties = new HashMap();
   }

   /**
    * Creates a JSONObject that is null.
    */
   public JSONObject( boolean isNull )
   {
      this();
      this.nullObject = isNull;
   }

   /**
    * Construct a JSONObject from a DynaBean.<br>
    * Assumes the object hierarchy is acyclical.
    *
    * @param dynaBean A DynaBean that can be used to initialize the contents of
    *        the JSONObject.
    * @deprecated use {@link fromObject} instead
    */
   public JSONObject( DynaBean dynaBean )
   {
      this( dynaBean, null, false );
   }

   /**
    * Construct a JSONObject from a subset of another JSONObject.
    *
    * @param jo A JSONObject.
    * @exception JSONException If a value is a non-finite number.
    * @deprecated use {@link fromObject} instead
    */
   public JSONObject( JSONObject jo )
   {
      this( jo, null, false );
   }

   /**
    * Construct a JSONObject from a Map.<br>
    * Assumes the object hierarchy is acyclical.
    *
    * @param map A map object that can be used to initialize the contents of the
    *        JSONObject.
    * @deprecated use {@link fromObject} instead
    */
   public JSONObject( Map map )
   {
      this( map, null, false );
   }

   /**
    * Construct a JSONObject from a string. This is the most commonly used
    * JSONObject constructor.
    *
    * @param string A string beginning with <code>{</code>&nbsp;<small>(left
    *        brace)</small> and ending with <code>}</code>&nbsp;<small>(right
    *        brace)</small>.
    * @exception JSONException If there is a syntax error in the source string.
    * @deprecated use {@link fromObject} instead
    */
   public JSONObject( String string )
   {
      this( new JSONTokener( string ), null, false );
   }

   /**
    * Construct a JSONObject from a DynaBean.<br>
    * Assumes the object hierarchy is acyclical.
    *
    * @param bean A DynaBean that can be used to initialize the contents of the
    *        JSONObject.
    */
   private JSONObject( DynaBean bean, String[] excludes, boolean ignoreDefaultExcludes )
   {
      this();

      if( bean == null ){
         this.nullObject = true;
         return;
      }

      DynaProperty[] props = bean.getDynaClass()
            .getDynaProperties();

      Collection exclusions = mergeExclusions( excludes, ignoreDefaultExcludes );
      for( int i = 0; i < props.length; i++ ){
         DynaProperty dynaProperty = props[i];
         String key = dynaProperty.getName();
         if( exclusions.contains( key ) ){
            continue;
         }
         Class type = dynaProperty.getType();
         Object value = bean.get( dynaProperty.getName() );
         setValue( this, key, value, type, excludes, ignoreDefaultExcludes );
      }
   }

   /**
    * Construct a JSONObject from a subset of another JSONObject.
    *
    * @param jo A JSONObject.
    * @exception JSONException If a value is a non-finite number.
    */
   private JSONObject( JSONObject jo, String[] excludes, boolean ignoreDefaultExcludes )
   {
      this();
      if( jo == null || jo.isNullObject() ){
         nullObject = true;
         return;
      }

      JSONArray sa = jo.names();
      Collection exclusions = mergeExclusions( excludes, ignoreDefaultExcludes );
      for( Iterator i = sa.iterator(); i.hasNext(); ){
         String key = (String) i.next();
         if( exclusions.contains( key ) ){
            continue;
         }
         putOpt( key, jo.opt( key ) );
      }
   }

   /**
    * Construct a JSONObject from a JSONTokener.
    *
    * @param x A JSONTokener object containing the source string.
    * @throws JSONException If there is a syntax error in the source string.
    */
   private JSONObject( JSONTokener x, String[] excludes, boolean ignoreDefaultExcludes )
   {
      this();
      char c;
      String key;

      if( x.matches( "null.*" ) ){
         this.nullObject = true;
         return;
      }

      if( x.nextClean() != '{' ){
         throw x.syntaxError( "A JSONObject text must begin with '{'" );
      }
      Collection exclusions = mergeExclusions( excludes, ignoreDefaultExcludes );
      for( ;; ){
         c = x.nextClean();
         switch( c )
         {
            case 0:
               throw x.syntaxError( "A JSONObject text must end with '}'" );
            case '}':
               return;
            default:
               x.back();
               key = x.nextValue( excludes, ignoreDefaultExcludes )
                     .toString();
         }

         /*
          * The key is followed by ':'. We will also tolerate '=' or '=>'.
          */

         c = x.nextClean();
         if( c == '=' ){
            if( x.next() != '>' ){
               x.back();
            }
         }else if( c != ':' ){
            throw x.syntaxError( "Expected a ':' after a key" );
         }
         Object v = x.nextValue();
         if( !JSONUtils.isFunctionHeader( v ) ){
            if( exclusions.contains( key ) ){
               switch( x.nextClean() )
               {
                  case ';':
                  case ',':
                     if( x.nextClean() == '}' ){
                        return;
                     }
                     x.back();
                     break;
                  case '}':
                     return;
                  default:
                     throw x.syntaxError( "Expected a ',' or '}'" );
               }
               continue;
            }
            if( v instanceof String && JSONUtils.mayBeJSON( (String) v ) ){
               put( key, JSONUtils.DOUBLE_QUOTE + v + JSONUtils.DOUBLE_QUOTE );
            }else{
               put( key, v, excludes, ignoreDefaultExcludes );
            }
         }else{
            // read params if any
            String params = JSONUtils.getFunctionParams( (String) v );
            // read function text
            int i = 0;
            StringBuffer sb = new StringBuffer();
            for( ;; ){
               char ch = x.next();
               if( ch == 0 ){
                  break;
               }
               if( ch == '{' ){
                  i++;
               }
               if( ch == '}' ){
                  i--;
               }
               sb.append( ch );
               if( i == 0 ){
                  break;
               }
            }
            if( i != 0 ){
               throw x.syntaxError( "Unbalanced '{' or '}' on prop: " + v );
            }
            // trim '{' at start and '}' at end
            String text = sb.toString();
            text = text.substring( 1, text.length() - 1 )
                  .trim();
            put( key, new JSONFunction( (params != null) ? StringUtils.split( params, "," ) : null,
                  text ) );
         }

         /*
          * Pairs are separated by ','. We will also tolerate ';'.
          */

         switch( x.nextClean() )
         {
            case ';':
            case ',':
               if( x.nextClean() == '}' ){
                  return;
               }
               x.back();
               break;
            case '}':
               return;
            default:
               throw x.syntaxError( "Expected a ',' or '}'" );
         }
      }
   }

   /**
    * Construct a JSONObject from a Map.<br>
    * Assumes the object hierarchy is acyclical.
    *
    * @param map A map object that can be used to initialize the contents of the
    *        JSONObject.
    */
   private JSONObject( Map map, String[] excludes, boolean ignoreDefaultExcludes )
   {
      if( map == null ){
         this.nullObject = true;
         return;
      }

      this.properties = new HashMap();
      Collection exclusions = mergeExclusions( excludes, ignoreDefaultExcludes );
      for( Iterator entries = map.entrySet()
            .iterator(); entries.hasNext(); ){
         Map.Entry entry = (Map.Entry) entries.next();
         Object k = entry.getKey();
         String key = (k instanceof String) ? (String) k : String.valueOf( k );
         if( exclusions.contains( key ) ){
            continue;
         }
         Object value = entry.getValue();
         if( value == null ){
            set( key, value );
         }else{
            setValue( this, key, value, value.getClass(), excludes, ignoreDefaultExcludes );
         }
      }
   }

   /**
    * Accumulate values under a key. It is similar to the put method except that
    * if there is already an object stored under the key then a JSONArray is
    * stored under the key to hold all of the accumulated values. If there is
    * already a JSONArray, then the new value is appended to it. In contrast,
    * the put method replaces the previous value.
    *
    * @param key A key string.
    * @param value An object to be accumulated under the key.
    * @return this.
    * @throws JSONException If the value is an invalid number or if the key is
    *         null.
    */
   public JSONObject accumulate( String key, Object value )
   {
      return accumulate( key, value, null, false );
   }

   /**
    * Accumulate values under a key. It is similar to the put method except that
    * if there is already an object stored under the key then a JSONArray is
    * stored under the key to hold all of the accumulated values. If there is
    * already a JSONArray, then the new value is appended to it. In contrast,
    * the put method replaces the previous value.
    *
    * @param key A key string.
    * @param value An object to be accumulated under the key.
    * @param excludes A group of property names to be excluded
    * @return this.
    * @throws JSONException If the value is an invalid number or if the key is
    *         null.
    */
   public JSONObject accumulate( String key, Object value, String[] excludes )
   {
      return accumulate( key, value, excludes, false );
   }

   /**
    * Accumulate values under a key. It is similar to the put method except that
    * if there is already an object stored under the key then a JSONArray is
    * stored under the key to hold all of the accumulated values. If there is
    * already a JSONArray, then the new value is appended to it. In contrast,
    * the put method replaces the previous value.
    *
    * @param key A key string.
    * @param value An object to be accumulated under the key.@param excludes A
    *        group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @return this.
    * @throws JSONException If the value is an invalid number or if the key is
    *         null.
    */
   public JSONObject accumulate( String key, Object value, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      if( isNullObject() ){
         throw new JSONException( "Can't accumulate on null object" );
      }

      JSONUtils.testValidity( value );
      Object o = opt( key );
      if( o == null ){
         set( key, value, excludes, ignoreDefaultExcludes );
      }else if( o instanceof JSONArray ){
         ((JSONArray) o).put( value, excludes, ignoreDefaultExcludes );
      }else{
         set( key, new JSONArray().put( o )
               .put( value, excludes, ignoreDefaultExcludes ) );
      }

      return this;
   }

   /**
    * Get the value object associated with a key.
    *
    * @param key A key string.
    * @return The object associated with the key.
    * @throws JSONException if the key is not found.
    */
   public Object get( String key )
   {
      verifyIsNull();
      Object o = opt( key );
      if( o == null ){
         throw new JSONException( "JSONObject[" + JSONUtils.quote( key ) + "] not found." );
      }
      return o;
   }

   /**
    * Get the boolean value associated with a key.
    *
    * @param key A key string.
    * @return The truth.
    * @throws JSONException if the value is not a Boolean or the String "true"
    *         or "false".
    */
   public boolean getBoolean( String key )
   {
      verifyIsNull();
      Object o = get( key );
      if( o.equals( Boolean.FALSE )
            || (o instanceof String && ((String) o).equalsIgnoreCase( "false" )) ){
         return false;
      }else if( o.equals( Boolean.TRUE )
            || (o instanceof String && ((String) o).equalsIgnoreCase( "true" )) ){
         return true;
      }
      throw new JSONException( "JSONObject[" + JSONUtils.quote( key ) + "] is not a Boolean." );
   }

   /**
    * Get the double value associated with a key.
    *
    * @param key A key string.
    * @return The numeric value.
    * @throws JSONException if the key is not found or if the value is not a
    *         Number object and cannot be converted to a number.
    */
   public double getDouble( String key )
   {
      verifyIsNull();
      Object o = get( key );
      try{
         return o instanceof Number ? ((Number) o).doubleValue() : Double.parseDouble( (String) o );
      }
      catch( Exception e ){
         throw new JSONException( "JSONObject[" + JSONUtils.quote( key ) + "] is not a number." );
      }
   }

   /**
    * Get the int value associated with a key. If the number value is too large
    * for an int, it will be clipped.
    *
    * @param key A key string.
    * @return The integer value.
    * @throws JSONException if the key is not found or if the value cannot be
    *         converted to an integer.
    */
   public int getInt( String key )
   {
      verifyIsNull();
      Object o = get( key );
      return o instanceof Number ? ((Number) o).intValue() : (int) getDouble( key );
   }

   /**
    * Get the JSONArray value associated with a key.
    *
    * @param key A key string.
    * @return A JSONArray which is the value.
    * @throws JSONException if the key is not found or if the value is not a
    *         JSONArray.
    */
   public JSONArray getJSONArray( String key )
   {
      verifyIsNull();
      Object o = get( key );
      if( o instanceof JSONArray ){
         return (JSONArray) o;
      }
      throw new JSONException( "JSONObject[" + JSONUtils.quote( key ) + "] is not a JSONArray." );
   }

   /**
    * Get the JSONObject value associated with a key.
    *
    * @param key A key string.
    * @return A JSONObject which is the value.
    * @throws JSONException if the key is not found or if the value is not a
    *         JSONObject.
    */
   public JSONObject getJSONObject( String key )
   {
      verifyIsNull();
      Object o = get( key );
      if( o instanceof JSONObject ){
         return (JSONObject) o;
      }
      throw new JSONException( "JSONObject[" + JSONUtils.quote( key ) + "] is not a JSONObject." );
   }

   /**
    * Get the long value associated with a key. If the number value is too long
    * for a long, it will be clipped.
    *
    * @param key A key string.
    * @return The long value.
    * @throws JSONException if the key is not found or if the value cannot be
    *         converted to a long.
    */
   public long getLong( String key )
   {
      verifyIsNull();
      Object o = get( key );
      return o instanceof Number ? ((Number) o).longValue() : (long) getDouble( key );
   }

   /**
    * Get the string associated with a key.
    *
    * @param key A key string.
    * @return A string which is the value.
    * @throws JSONException if the key is not found.
    */
   public String getString( String key )
   {
      verifyIsNull();
      return get( key ).toString();
   }

   /**
    * Determine if the JSONObject contains a specific key.
    *
    * @param key A key string.
    * @return true if the key exists in the JSONObject.
    */
   public boolean has( String key )
   {
      verifyIsNull();
      return this.properties.containsKey( key );
   }

   public boolean isArray()
   {
      return false;
   }

   public boolean isEmpty()
   {
      return this.properties.isEmpty();
   }

   /**
    * Returs if this object is a null JSONObject.
    */
   public boolean isNullObject()
   {
      return nullObject;
   }

   /**
    * Get an enumeration of the keys of the JSONObject.
    *
    * @return An iterator of the keys.
    */
   public Iterator keys()
   {
      verifyIsNull();
      return this.properties.keySet()
            .iterator();
   }

   /**
    * Get the number of keys stored in the JSONObject.
    *
    * @return The number of keys in the JSONObject.
    */
   public int length()
   {
      verifyIsNull();
      return this.properties.size();
   }

   /**
    * Produce a JSONArray containing the names of the elements of this
    * JSONObject.
    *
    * @return A JSONArray containing the key strings, or null if the JSONObject
    *         is empty.
    */
   public JSONArray names()
   {
      verifyIsNull();
      JSONArray ja = new JSONArray();
      Iterator keys = keys();
      while( keys.hasNext() ){
         ja.put( keys.next() );
      }
      return ja;
   }

   /**
    * Get an optional value associated with a key.
    *
    * @param key A key string.
    * @return An object which is the value, or null if there is no value.
    */
   public Object opt( String key )
   {
      verifyIsNull();
      return key == null ? null : this.properties.get( key );
   }

   /**
    * Get an optional boolean associated with a key. It returns false if there
    * is no such key, or if the value is not Boolean.TRUE or the String "true".
    *
    * @param key A key string.
    * @return The truth.
    */
   public boolean optBoolean( String key )
   {
      verifyIsNull();
      return optBoolean( key, false );
   }

   /**
    * Get an optional boolean associated with a key. It returns the defaultValue
    * if there is no such key, or if it is not a Boolean or the String "true" or
    * "false" (case insensitive).
    *
    * @param key A key string.
    * @param defaultValue The default.
    * @return The truth.
    */
   public boolean optBoolean( String key, boolean defaultValue )
   {
      verifyIsNull();
      try{
         return getBoolean( key );
      }
      catch( Exception e ){
         return defaultValue;
      }
   }

   /**
    * Get an optional double associated with a key, or NaN if there is no such
    * key or if its value is not a number. If the value is a string, an attempt
    * will be made to evaluate it as a number.
    *
    * @param key A string which is the key.
    * @return An object which is the value.
    */
   public double optDouble( String key )
   {
      verifyIsNull();
      return optDouble( key, Double.NaN );
   }

   /**
    * Get an optional double associated with a key, or the defaultValue if there
    * is no such key or if its value is not a number. If the value is a string,
    * an attempt will be made to evaluate it as a number.
    *
    * @param key A key string.
    * @param defaultValue The default.
    * @return An object which is the value.
    */
   public double optDouble( String key, double defaultValue )
   {
      verifyIsNull();
      try{
         Object o = opt( key );
         return o instanceof Number ? ((Number) o).doubleValue()
               : new Double( (String) o ).doubleValue();
      }
      catch( Exception e ){
         return defaultValue;
      }
   }

   /**
    * Get an optional int value associated with a key, or zero if there is no
    * such key or if the value is not a number. If the value is a string, an
    * attempt will be made to evaluate it as a number.
    *
    * @param key A key string.
    * @return An object which is the value.
    */
   public int optInt( String key )
   {
      verifyIsNull();
      return optInt( key, 0 );
   }

   /**
    * Get an optional int value associated with a key, or the default if there
    * is no such key or if the value is not a number. If the value is a string,
    * an attempt will be made to evaluate it as a number.
    *
    * @param key A key string.
    * @param defaultValue The default.
    * @return An object which is the value.
    */
   public int optInt( String key, int defaultValue )
   {
      verifyIsNull();
      try{
         return getInt( key );
      }
      catch( Exception e ){
         return defaultValue;
      }
   }

   /**
    * Get an optional JSONArray associated with a key. It returns null if there
    * is no such key, or if its value is not a JSONArray.
    *
    * @param key A key string.
    * @return A JSONArray which is the value.
    */
   public JSONArray optJSONArray( String key )
   {
      verifyIsNull();
      Object o = opt( key );
      return o instanceof JSONArray ? (JSONArray) o : null;
   }

   /**
    * Get an optional JSONObject associated with a key. It returns null if there
    * is no such key, or if its value is not a JSONObject.
    *
    * @param key A key string.
    * @return A JSONObject which is the value.
    */
   public JSONObject optJSONObject( String key )
   {
      verifyIsNull();
      Object o = opt( key );
      return o instanceof JSONObject ? (JSONObject) o : null;
   }

   /**
    * Get an optional long value associated with a key, or zero if there is no
    * such key or if the value is not a number. If the value is a string, an
    * attempt will be made to evaluate it as a number.
    *
    * @param key A key string.
    * @return An object which is the value.
    */
   public long optLong( String key )
   {
      verifyIsNull();
      return optLong( key, 0 );
   }

   /**
    * Get an optional long value associated with a key, or the default if there
    * is no such key or if the value is not a number. If the value is a string,
    * an attempt will be made to evaluate it as a number.
    *
    * @param key A key string.
    * @param defaultValue The default.
    * @return An object which is the value.
    */
   public long optLong( String key, long defaultValue )
   {
      verifyIsNull();
      try{
         return getLong( key );
      }
      catch( Exception e ){
         return defaultValue;
      }
   }

   /**
    * Get an optional string associated with a key. It returns an empty string
    * if there is no such key. If the value is not a string and is not null,
    * then it is coverted to a string.
    *
    * @param key A key string.
    * @return A string which is the value.
    */
   public String optString( String key )
   {
      verifyIsNull();
      return optString( key, "" );
   }

   /**
    * Get an optional string associated with a key. It returns the defaultValue
    * if there is no such key.
    *
    * @param key A key string.
    * @param defaultValue The default.
    * @return A string which is the value.
    */
   public String optString( String key, String defaultValue )
   {
      verifyIsNull();
      Object o = opt( key );
      return o != null ? o.toString() : defaultValue;
   }

   /**
    * Put a key/boolean pair in the JSONObject.
    *
    * @param key A key string.
    * @param value A boolean which is the value.
    * @return this.
    * @throws JSONException If the key is null.
    */
   public JSONObject put( String key, boolean value )
   {
      verifyIsNull();
      put( key, value ? Boolean.TRUE : Boolean.FALSE );
      return this;
   }

   /**
    * Put a key/value pair in the JSONObject, where the value will be a
    * JSONArray which is produced from a Collection.
    *
    * @param key A key string.
    * @param value A Collection value.
    * @return this.
    * @throws JSONException
    */
   public JSONObject put( String key, Collection value )
   {
      return put( key, value, null, false );
   }

   /**
    * Put a key/value pair in the JSONObject, where the value will be a
    * JSONArray which is produced from a Collection.
    *
    * @param key A key string.
    * @param value A Collection value.
    * @param excludes A group of property names to be excluded
    * @return this.
    * @throws JSONException
    */
   public JSONObject put( String key, Collection value, String[] excludes )
   {
      return put( key, value, excludes, false );
   }

   /**
    * Put a key/value pair in the JSONObject, where the value will be a
    * JSONArray which is produced from a Collection.
    *
    * @param key A key string.
    * @param value A Collection value.
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @return this.
    * @throws JSONException
    */
   public JSONObject put( String key, Collection value, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      verifyIsNull();
      put( key, JSONArray.fromObject( value, excludes, ignoreDefaultExcludes ) );
      return this;
   }

   /**
    * Put a key/double pair in the JSONObject.
    *
    * @param key A key string.
    * @param value A double which is the value.
    * @return this.
    * @throws JSONException If the key is null or if the number is invalid.
    */
   public JSONObject put( String key, double value )
   {
      verifyIsNull();
      Double d = new Double( value );
      JSONUtils.testValidity( d );
      put( key, d );
      return this;
   }

   /**
    * Put a key/int pair in the JSONObject.
    *
    * @param key A key string.
    * @param value An int which is the value.
    * @return this.
    * @throws JSONException If the key is null.
    */
   public JSONObject put( String key, int value )
   {
      verifyIsNull();
      put( key, new Integer( value ) );
      return this;
   }

   /**
    * Put a key/long pair in the JSONObject.
    *
    * @param key A key string.
    * @param value A long which is the value.
    * @return this.
    * @throws JSONException If the key is null.
    */
   public JSONObject put( String key, long value )
   {
      verifyIsNull();
      put( key, new Long( value ) );
      return this;
   }

   /**
    * Put a key/value pair in the JSONObject, where the value will be a
    * JSONObject which is produced from a Map.
    *
    * @param key A key string.
    * @param value A Map value.
    * @return this.
    * @throws JSONException
    */
   public JSONObject put( String key, Map value )
   {
      return put( key, value, null, false );
   }

   /**
    * Put a key/value pair in the JSONObject, where the value will be a
    * JSONObject which is produced from a Map.
    *
    * @param key A key string.
    * @param value A Map value.
    * @param excludes A group of property names to be excluded
    * @return this.
    * @throws JSONException
    */
   public JSONObject put( String key, Map value, String[] excludes )
   {
      return put( key, value, excludes, false );
   }

   /**
    * Put a key/value pair in the JSONObject, where the value will be a
    * JSONObject which is produced from a Map.
    *
    * @param key A key string.
    * @param value A Map value.
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @return this.
    * @throws JSONException
    */
   public JSONObject put( String key, Map value, String[] excludes, boolean ignoreDefaultExcludes )
   {
      verifyIsNull();
      put( key, JSONObject.fromObject( value, excludes, ignoreDefaultExcludes ) );
      return this;
   }

   /**
    * Put a key/value pair in the JSONObject. If the value is null, then the key
    * will be removed from the JSONObject if it is present.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object.
    * @return this.
    * @throws JSONException If the value is non-finite number or if the key is
    *         null.
    */
   public JSONObject put( String key, Object value )
   {
      return put( key, value, null, false );
   }

   /**
    * Put a key/value pair in the JSONObject. If the value is null, then the key
    * will be removed from the JSONObject if it is present.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object.
    * @param excludes A group of property names to be excluded
    * @return this.
    * @throws JSONException If the value is non-finite number or if the key is
    *         null.
    */
   public JSONObject put( String key, Object value, String[] excludes )
   {
      return put( key, value, excludes, false );
   }

   /**
    * Put a key/value pair in the JSONObject. If the value is null, then the key
    * will be removed from the JSONObject if it is present.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object.
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @return this.
    * @throws JSONException If the value is non-finite number or if the key is
    *         null.
    */
   public JSONObject put( String key, Object value, String[] excludes, boolean ignoreDefaultExcludes )
   {
      verifyIsNull();
      if( key == null ){
         throw new JSONException( "Null key." );
      }
      if( value != null ){
         if( this.properties.containsKey( key ) ){
            accumulate( key, value, excludes, ignoreDefaultExcludes );
            getJSONArray( key ).setExpandElements( true );
         }else{
            set( key, value, excludes, ignoreDefaultExcludes );
         }
      }else{
         remove( key );
      }
      return this;
   }

   /**
    * Put a key/value pair in the JSONObject, but only if the key and the value
    * are both non-null.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object.
    * @return this.
    * @throws JSONException If the value is a non-finite number.
    */
   public JSONObject putOpt( String key, Object value )
   {
      return put( key, value, null, false );
   }

   /**
    * Put a key/value pair in the JSONObject, but only if the key and the value
    * are both non-null.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object. *
    * @param excludes A group of property names to be excluded
    * @return this.
    * @throws JSONException If the value is a non-finite number.
    */
   public JSONObject putOpt( String key, Object value, String[] excludes )
   {
      return put( key, value, excludes, false );
   }

   /**
    * Put a key/value pair in the JSONObject, but only if the key and the value
    * are both non-null.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object.
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @return this.
    * @throws JSONException If the value is a non-finite number.
    */
   public JSONObject putOpt( String key, Object value, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      verifyIsNull();
      if( key != null && value != null ){
         put( key, value );
      }
      return this;
   }

   /**
    * Remove a name and its value, if present.
    *
    * @param key The name to be removed.
    * @return The value that was associated with the name, or null if there was
    *         no value.
    */
   public Object remove( String key )
   {
      verifyIsNull();
      return this.properties.remove( key );
   }

   /**
    * Produce a JSONArray containing the values of the members of this
    * JSONObject.
    *
    * @param names A JSONArray containing a list of key strings. This determines
    *        the sequence of the values in the result.
    * @return A JSONArray of values.
    * @throws JSONException If any of the values are non-finite numbers.
    */
   public JSONArray toJSONArray( JSONArray names )
   {
      verifyIsNull();
      if( names == null || names.length() == 0 ){
         return null;
      }
      JSONArray ja = new JSONArray();
      for( int i = 0; i < names.length(); i += 1 ){
         ja.put( this.opt( names.getString( i ) ) );
      }
      return ja;
   }

   /**
    * Make a JSON text of this JSONObject. For compactness, no whitespace is
    * added. If this would not result in a syntactically correct JSON text, then
    * null will be returned instead.
    * <p>
    * Warning: This method assumes that the data structure is acyclical.
    *
    * @return a printable, displayable, portable, transmittable representation
    *         of the object, beginning with <code>{</code>&nbsp;<small>(left
    *         brace)</small> and ending with <code>}</code>&nbsp;<small>(right
    *         brace)</small>.
    */
   public String toString()
   {
      if( isNullObject() ){
         return JSONNull.getInstance()
               .toString();
      }
      try{
         Iterator keys = keys();
         StringBuffer sb = new StringBuffer( "{" );

         while( keys.hasNext() ){
            if( sb.length() > 1 ){
               sb.append( ',' );
            }
            Object o = keys.next();
            sb.append( JSONUtils.quote( o.toString() ) );
            sb.append( ':' );
            sb.append( JSONUtils.valueToString( this.properties.get( o ) ) );
         }
         sb.append( '}' );
         return sb.toString();
      }
      catch( Exception e ){
         return null;
      }
   }

   /**
    * Make a prettyprinted JSON text of this JSONObject.
    * <p>
    * Warning: This method assumes that the data structure is acyclical.
    *
    * @param indentFactor The number of spaces to add to each level of
    *        indentation.
    * @return a printable, displayable, portable, transmittable representation
    *         of the object, beginning with <code>{</code>&nbsp;<small>(left
    *         brace)</small> and ending with <code>}</code>&nbsp;<small>(right
    *         brace)</small>.
    * @throws JSONException If the object contains an invalid number.
    */
   public String toString( int indentFactor )
   {
      if( isNullObject() ){
         return JSONNull.getInstance()
               .toString();
      }
      return toString( indentFactor, 0 );
   }

   /**
    * Make a prettyprinted JSON text of this JSONObject.
    * <p>
    * Warning: This method assumes that the data structure is acyclical.
    *
    * @param indentFactor The number of spaces to add to each level of
    *        indentation.
    * @param indent The indentation of the top level.
    * @return a printable, displayable, transmittable representation of the
    *         object, beginning with <code>{</code>&nbsp;<small>(left brace)</small>
    *         and ending with <code>}</code>&nbsp;<small>(right brace)</small>.
    * @throws JSONException If the object contains an invalid number.
    */
   public String toString( int indentFactor, int indent )
   {
      if( isNullObject() ){
         return JSONNull.getInstance()
               .toString();
      }
      int i;
      int n = length();
      if( n == 0 ){
         return "{}";
      }
      Iterator keys = keys();
      StringBuffer sb = new StringBuffer( "{" );
      int newindent = indent + indentFactor;
      Object o;
      if( n == 1 ){
         o = keys.next();
         sb.append( JSONUtils.quote( o.toString() ) );
         sb.append( ": " );
         sb.append( JSONUtils.valueToString( this.properties.get( o ), indentFactor, indent ) );
      }else{
         while( keys.hasNext() ){
            o = keys.next();
            if( sb.length() > 1 ){
               sb.append( ",\n" );
            }else{
               sb.append( '\n' );
            }
            for( i = 0; i < newindent; i += 1 ){
               sb.append( ' ' );
            }
            sb.append( JSONUtils.quote( o.toString() ) );
            sb.append( ": " );
            sb.append( JSONUtils.valueToString( this.properties.get( o ), indentFactor, newindent ) );
         }
         if( sb.length() > 1 ){
            sb.append( '\n' );
            for( i = 0; i < indent; i += 1 ){
               sb.append( ' ' );
            }
         }
         for( i = 0; i < indent; i += 1 ){
            sb.insert( 0, ' ' );
         }
      }
      sb.append( '}' );
      return sb.toString();
   }

   /**
    * Write the contents of the JSONObject as JSON text to a writer. For
    * compactness, no whitespace is added.
    * <p>
    * Warning: This method assumes that the data structure is acyclical.
    *
    * @return The writer.
    * @throws JSONException
    */
   public Writer write( Writer writer )
   {
      try{
         if( isNullObject() ){
            writer.write( JSONNull.getInstance()
                  .toString() );
            return writer;
         }

         boolean b = false;
         Iterator keys = keys();
         writer.write( '{' );

         while( keys.hasNext() ){
            if( b ){
               writer.write( ',' );
            }
            Object k = keys.next();
            writer.write( JSONUtils.quote( k.toString() ) );
            writer.write( ':' );
            Object v = this.properties.get( k );
            if( v instanceof JSONObject ){
               ((JSONObject) v).write( writer );
            }else if( v instanceof JSONArray ){
               ((JSONArray) v).write( writer );
            }else{
               writer.write( JSONUtils.valueToString( v ) );
            }
            b = true;
         }
         writer.write( '}' );
         return writer;
      }
      catch( IOException e ){
         throw new JSONException( e );
      }
   }

   /**
    * Put a key/value pair in the JSONObject.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object.
    * @return this.
    * @throws JSONException If the value is non-finite number or if the key is
    *         null.
    */
   private JSONObject set( String key, Object value )
   {
      return set( key, value, null, false );
   }

   /**
    * Put a key/value pair in the JSONObject.
    *
    * @param key A key string.
    * @param value An object which is the value. It should be of one of these
    *        types: Boolean, Double, Integer, JSONArray, JSONObject, Long,
    *        String, or the JSONNull object.}
    * @param excludes A group of property names to be excluded
    * @param ignoreDefaultExcludes A flag for ignoring the default exclusions of
    *        property names
    * @return this.
    * @throws JSONException If the value is non-finite number or if the key is
    *         null.
    */
   private JSONObject set( String key, Object value, String[] excludes,
         boolean ignoreDefaultExcludes )
   {
      verifyIsNull();
      if( key == null ){
         throw new JSONException( "Null key." );
      }

      if( (value != null && Class.class.isAssignableFrom( value.getClass() ))
            || value instanceof Class ){
         this.properties.put( key, ((Class) value).getName() );
      }else if( JSONUtils.isFunction( value ) ){
         this.properties.put( key, value );
      }else if( value instanceof JSONString ){
         this.properties.put( key, JSONSerializer.toJSON( (JSONString) value, excludes,
               ignoreDefaultExcludes ) );
      }else if( JSONUtils.isArray( value ) ){
         this.properties.put( key, JSONArray.fromObject( value, excludes, ignoreDefaultExcludes ) );
      }else if( value instanceof JSON ){
         this.properties.put( key, value );
      }else if( JSONUtils.isString( value ) ){
         String str = String.valueOf( value );
         if( JSONUtils.mayBeJSON( str ) ){
            try{
               this.properties.put( key, JSONSerializer.toJSON( str, excludes,
                     ignoreDefaultExcludes ) );
            }
            catch( JSONException jsone ){
               this.properties.put( key, JSONUtils.stripQuotes( str ) );
            }
         }else{
            if( value == null ){
               this.properties.put( key, "" );
            }else{
               this.properties.put( key, JSONUtils.stripQuotes( str ) );
            }
         }
      }else if( JSONUtils.isNumber( value ) ){
         JSONUtils.testValidity( value );
         this.properties.put( key, JSONUtils.transformNumber( (Number) value ) );
      }else if( JSONUtils.isBoolean( value ) ){
         this.properties.put( key, value );
      }else{
         this.properties.put( key, fromObject( value, excludes, ignoreDefaultExcludes ) );
      }

      return this;
   }

   /**
    * Checks if this object is a "null" object.
    */
   private void verifyIsNull()
   {
      if( isNullObject() ){
         throw new JSONException( "null object" );
      }
   }
}